/* global window, setTimeout, clearTimeout */
import {isBrowser, log} from '../utils';
import {getPageLoadPromise, resizeDrawingBuffer} from '../webgl-utils';
import {createGLContext, isWebGL, resetParameters} from '../webgl';
import {Framebuffer} from '../webgl';

// Node.js polyfills for requestAnimationFrame and cancelAnimationFrame
export function requestAnimationFrame(callback) {
  return isBrowser ? window.requestAnimationFrame(callback) : setTimeout(callback, 1000 / 60);
}

export function cancelAnimationFrame(timerId) {
  return isBrowser ? window.cancelAnimationFrame(timerId) : clearTimeout(timerId);
}

const DEFAULT_GL_OPTIONS = {
  preserveDrawingBuffer: true
};

export default class AnimationLoop {
  /*
   * @param {HTMLCanvasElement} canvas - if provided, width and height will be passed to context
   */
  constructor(props = {}) {
    const {
      onCreateContext = opts => createGLContext(opts),
      onInitialize = () => {},
      onRender = () => {},
      onFinalize = () => {},

      gl = null,
      glOptions = {},
      width = null,
      height = null,

      createFramebuffer = false,

      // view parameters
      autoResizeViewport = true,
      autoResizeDrawingBuffer = true
    } = props;

    let {
      useDevicePixels = true
    } = props;

    if ('useDevicePixelRatio' in props) {
      log.deprecated('useDevicePixelRatio', 'useDevicePixels');
      useDevicePixels = props.useDevicePixelRatio;
    }

    this._onCreateContext = onCreateContext;
    this.start = this.start.bind(this);
    this.stop = this.stop.bind(this);
    this._renderFrame = this._renderFrame.bind(this);

    this.glOptions = glOptions;
    this._createFramebuffer = createFramebuffer;

    this._onInitialize = onInitialize;
    this._onRender = onRender;
    this._onFinalize = onFinalize;

    this.width = width;
    this.height = height;

    this.gl = gl;

    this.setProps({
      autoResizeViewport,
      autoResizeDrawingBuffer,
      useDevicePixels
    });

    return this;
  }

  setNeedsRedraw(reason) {
    this.needsRedraw = this.needsRedraw || reason;
    return this;
  }

  setProps(props) {
    if ('autoResizeViewport' in props) {
      this.autoResizeViewport = props.autoResizeViewport;
    }
    if ('autoResizeDrawingBuffer' in props) {
      this.autoResizeDrawingBuffer = props.autoResizeDrawingBuffer;
    }
    if ('useDevicePixels' in props) {
      this.useDevicePixels = props.useDevicePixels;
    }
    return this;
  }

  // Starts a render loop if not already running
  // @param {Object} context - contains frame specific info (E.g. tick, width, height, etc)
  start(opts = {}) {
    this._stopped = false;
    // console.debug(`Starting ${this.constructor.name}`);
    if (!this._animationFrameId) {
      // Wait for start promise before rendering frame
      this._startPromise = getPageLoadPromise()
      .then(() => {
        if (this._stopped) {
          return null;
        }

        // Create the WebGL context
        this._createWebGLContext(opts);

        // Initialize the callback data
        this._initializeCallbackData();
        this._updateCallbackData();

        // Default viewport setup, in case onInitialize wants to render
        this._resizeCanvasDrawingBuffer();
        this._resizeViewport();

        // Note: onIntialize can return a promise (in case it needs to load resources)
        return this._onInitialize(this._callbackData);
      })
      .then(appContext => {
        if (!this._stopped) {
          this._addCallbackData(appContext || {});
          if (appContext !== false && !this._animationFrameId) {
            this._animationFrameId = requestAnimationFrame(this._renderFrame);
          }
        }
      });

    }
    return this;
  }

  // Stops a render loop if already running, finalizing
  stop() {
    // console.debug(`Stopping ${this.constructor.name}`);
    if (this._animationFrameId) {
      this._finalizeCallbackData();
      cancelAnimationFrame(this._animationFrameId);
      this._animationFrameId = null;
      this._stopped = true;
    }
    return this;
  }

  // DEPRECATED METHODS

  // Update parameters
  setViewParameters({
    autoResizeDrawingBuffer = true,
    autoResizeCanvas = true,
    autoResizeViewport = true,
    useDevicePixels = true,
    useDevicePixelRatio = null // deprecated
  }) {
    log.deprecated('AnimationLoop.setViewParameters', 'AnimationLoop.setProps');
    this.autoResizeViewport = autoResizeViewport;
    this.autoResizeCanvas = autoResizeCanvas;
    this.autoResizeDrawingBuffer = autoResizeDrawingBuffer;
    this.useDevicePixels = useDevicePixels;
    if (useDevicePixelRatio !== null) {
      log.deprecated('useDevicePixelRatio', 'useDevicePixels');
      this.useDevicePixels = useDevicePixelRatio;
    }
    return this;
  }

  // PRIVATE METHODS

  _setupFrame() {
    if (this._onSetupFrame) {
      // call callback
      this._onSetupFrame(this._callbackData);
      // end callback
    } else {
      this._resizeCanvasDrawingBuffer();
      this._resizeViewport();
      this._resizeFramebuffer();
    }
  }

  /**
   * @private
   * Handles a render loop frame- updates context and calls the application
   * callback
   */
  _renderFrame() {
    this._setupFrame();
    this._updateCallbackData();

    // call callback
    this._onRender(this._callbackData);
    // end callback

    // Increment tick
    this._callbackData.tick++;

    if (!this._stopped) {
      // Request another render frame (now )
      this._animationFrameId = requestAnimationFrame(this._renderFrame);
    }
  }

  // Initialize the  object that will be passed to app callbacks
  _initializeCallbackData() {
    this._callbackData = {
      gl: this.gl,
      canvas: this.gl.canvas,
      framebuffer: this.framebuffer,
      stop: this.stop,
      // Initial values
      tick: 0,
      tock: 0,
      useDevicePixels: this.useDevicePixels
    };
  }

  // Update the context object that will be passed to app callbacks
  _updateCallbackData() {
    // CallbackData width and height represent drawing buffer width and height
    const {canvas} = this.gl;
    this._callbackData.width = canvas.width;
    this._callbackData.height = canvas.height;
    this._callbackData.aspect = canvas.width / canvas.height;
  }

  _finalizeCallbackData() {
    // call callback
    this._onFinalize(this._callbackData);
    // end callback
  }

  // Add application's data to the app context object
  _addCallbackData(appContext) {
    if (typeof appContext === 'object' && appContext !== null) {
      this._callbackData = Object.assign({}, this._callbackData, appContext);
    }
  }

  // Either uses supplied or existing context, or calls provided callback to create one
  _createWebGLContext(opts) {
    // Create the WebGL context if necessary
    opts = Object.assign({}, opts, DEFAULT_GL_OPTIONS, this.glOptions);
    if (opts.gl) {
      this.gl = opts.gl;
    } else {
      this.gl = this._onCreateContext(opts);
    }
    if (!isWebGL(this.gl)) {
      throw new Error('AnimationLoop.onCreateContext - illegal context returned');
    }

    // Setup default framebuffer
    if (this._createFramebuffer) {
      this.framebuffer = new Framebuffer(this.gl);
    }
    // Reset the WebGL context.
    resetParameters(this.gl);
  }

  // Default viewport setup
  _resizeViewport() {
    if (this.autoResizeViewport) {
      this.gl.viewport(0, 0, this.gl.canvas.width, this.gl.canvas.height);
    }
  }

  _resizeFramebuffer() {
    if (this.framebuffer) {
      this.framebuffer.resize({width: this.gl.canvas.width, height: this.gl.canvas.height});
    }
  }

  // Resize the render buffer of the canvas to match canvas client size
  // Optionally multiplying with devicePixel ratio
  _resizeCanvasDrawingBuffer() {
    if (this.autoResizeDrawingBuffer) {
      resizeDrawingBuffer(this.gl.canvas, {useDevicePixels: this.useDevicePixels});
    }
  }
}
